import six
import requests

from requests.auth import HTTPBasicAuth

from txclib import utils
from txclib.urls import API_URLS, HOSTNAMES
from txclib.log import logger


class Api(object):

    USERNAME = 'api'
    VALID_CALLS = ['auth_check', 'projects', 'organizations', 'formats']

    def map_paths_to_hostnames(cls, path_to_tx, host):
        domains = utils.get_api_domains(path_to_tx, host)
        return {
            path: domains[key] for key, paths in HOSTNAMES.items()
            for path in paths
        }

    def __init__(self, token=None, username=None, password=None,
                 path_to_tx=None, host=None):
        self.hostnames = self.map_paths_to_hostnames(path_to_tx, host)
        if token:
            self.token = token
            self.username = self.USERNAME
        elif username and password:
            self.token = password
            self.username = username
        else:
            logger.error("Authorization credentials are missing. Make sure "
                         "that you have run `tx init` to setup your "
                         "credentials.")

    def get(self, api_call, *args, **kwargs):
        """
        Performs the GET API call specified by api_call and
        parses the response
        """
        # mock response
        if api_call not in self.VALID_CALLS:
            raise Exception(
                "Tried to perform unsupported API call {}".format(
                    api_call
                )
            )

        hostname = self.hostnames[api_call]
        url = API_URLS[api_call] % kwargs
        url = "{}{}".format(hostname, url)

        try:
            response = requests.get(
                url, auth=HTTPBasicAuth(self.username, self.token)
            )
            response.raise_for_status()
            all_data = response.json()
        except Exception as e:
            logger.debug(six.u(str(e)))
            raise

        next_page = response.links.get('next')
        while next_page:
            try:
                response = requests.get(
                    next_page['url'],
                    auth=HTTPBasicAuth(self.USERNAME, self.token)
                )
                response.raise_for_status()
                all_data.extend(response.json())
                next_page = response.links.get('next')
            except Exception as e:
                logger.debug(six.u(str(e)))
                raise
        return all_data



import fileinput
import os
import shutil
import unittest
import sys
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from mock import patch, MagicMock, call
from six import assertRaisesRegex
from txclib.commands import _set_source_file, _set_translation, cmd_pull, \
    cmd_init, cmd_config, cmd_status, cmd_help, UnInitializedError
from txclib.cmdline import main
from txclib.parsers import MAPPING, MAPPINGREMOTE, MAPPINGBULK


class TestCommands(unittest.TestCase):
    def test_cmd_pull_return_exception_when_dir_not_initialized(self):
        """Test when tx is not instantiated, that proper error is thrown"""
        with self.assertRaises(UnInitializedError):
            cmd_pull([], None)

    def test_set_source_file_when_dir_not_initialized(self):
        with self.assertRaises(UnInitializedError):
            _set_source_file(path_to_tx=None, resource='dummy_resource.en',
                             lang='en', path_to_file='dummy')

    def test_set_translation_when_dir_not_initialized(self):
        with self.assertRaises(UnInitializedError):
            _set_translation(path_to_tx=None, resource="dummy_resource.md",
                             lang='en', path_to_file='invalid')


class TestStatusCommand(unittest.TestCase):
    @patch('txclib.commands.project')
    def test_status(self, mock_p):
        """Test status command"""
        mock_project = MagicMock()
        mock_project.get_chosen_resources.return_value = ['foo.bar']
        mock_project.get_resource_files.return_value = {
            "fr": "translations/foo.bar/fr.po",
            "de": "translations/foo.bar/de.po"
        }
        mock_p.Project.return_value = mock_project
        cmd_status([], None)
        mock_project.get_chosen_resources.assert_called_once_with([])
        self.assertEqual(mock_project.get_resource_files.call_count, 1)


class TestHelpCommand(unittest.TestCase):
    def test_help(self):
        out = StringIO()
        sys.stdout = out
        cmd_help([], None)
        output = out.getvalue().strip()
        self.assertTrue(
            all(
                c in output for c in
                ['delete', 'help', 'init', 'pull', 'push', 'set', 'status']
            )
        )

        # call for specific command
        with patch('txclib.commands.cmd_pull', spec=cmd_pull) as pull_mock:
            cmd_help(['pull'], None)
            pull_mock.assert_called_once_with(['--help'], None)


class TestInitCommand(unittest.TestCase):

    def setUp(self):
        self.curr_dir = os.getcwd()
        self.config_file = '.tx/config'
        os.chdir('./tests/project_dir/')

    def tearDown(self, *args, **kwargs):
        shutil.rmtree('.tx', ignore_errors=False, onerror=None)
        os.chdir(self.curr_dir)
        super(TestInitCommand, self).tearDown(*args, **kwargs)

    def test_init(self):
        argv = []
        config_text = "[main]\nhost = https://www.transifex.com\n\n"
        with patch('txclib.commands.project.Project'):
            with patch('txclib.commands.cmd_config') as set_mock:
                cmd_init(argv, '')
                set_mock.assert_called_once_with([], os.getcwd())
        self.assertTrue(os.path.exists('./.tx'))
        self.assertTrue(os.path.exists('./.tx/config'))
        self.assertEqual(open(self.config_file).read(), config_text)

    def test_init_skipsetup(self):
        argv = ['--skipsetup']
        with patch('txclib.commands.project.Project'):
            with patch('txclib.commands.cmd_config') as set_mock:
                cmd_init(argv, '')
                self.assertEqual(set_mock.call_count, 0)
        self.assertTrue(os.path.exists('./.tx'))
        self.assertTrue(os.path.exists('./.tx/config'))

    @patch('txclib.commands.utils.confirm')
    def test_init_save_N(self, confirm_mock):
        os.mkdir('./.tx')
        open('./.tx/config', 'a').close()
        argv = []
        confirm_mock.return_value = False
        with patch('txclib.commands.project.Project') as project_mock:
                cmd_init(argv, '')
                self.assertEqual(project_mock.call_count, 0)
        self.assertTrue(os.path.exists('./.tx'))
        self.assertEqual(confirm_mock.call_count, 1)

    @patch('txclib.commands.utils.confirm')
    def test_init_save_y(self, confirm_mock):
        os.mkdir('./.tx')
        open('./.tx/config', 'a').close()
        argv = []
        confirm_mock.return_value = True
        with patch('txclib.commands.project.Project'):
            with patch('txclib.commands.cmd_config') as set_mock:
                cmd_init(argv, '')
                set_mock.assert_called()
        self.assertTrue(os.path.exists('./.tx'))
        self.assertEqual(confirm_mock.call_count, 1)

    def test_init_force_save(self):
        os.mkdir('./.tx')
        argv = ['--force-save']
        with patch('txclib.commands.project.Project'):
            with patch('txclib.commands.cmd_config') as set_mock:
                cmd_init(argv, '')
                set_mock.assert_called()
        self.assertTrue(os.path.exists('./.tx'))
        self.assertTrue(os.path.exists('./.tx/config'))


class TestPullCommand(unittest.TestCase):
    @patch('txclib.utils.get_current_branch')
    @patch('txclib.commands.logger')
    @patch('txclib.commands.project.Project')
    def test_pull_with_branch_no_git_repo(self, project_mock, log_mock, bmock):
        bmock.return_value = None
        pr_instance = MagicMock()
        project_mock.return_value = pr_instance
        with self.assertRaises(SystemExit):
            cmd_pull(['--branch'], '.')
        log_mock.error.assert_called_once_with(
            "You specified the --branch option but current "
            "directory does not seem to belong in any git repo.")
        self.assertEqual(pr_instance.pull.call_count, 0)

    @patch('txclib.utils.get_current_branch')
    @patch('txclib.commands.logger')
    @patch('txclib.commands.project.Project')
    def test_pull_branch_git_repo(self, project_mock, log_mock, bmock):
        bmock.return_value = 'a-branch'
        pr_instance = MagicMock()
        project_mock.return_value = pr_instance
        cmd_pull(['--branch'], '.')
        self.assertEqual(pr_instance.pull.call_count, 1)
        pull_call = call(
            branch='a-branch', fetchall=False, fetchsource=False,
            force=False, languages=[], minimum_perc=None, mode=None,
            overwrite=True, pseudo=False, resources=[], skip=False,
            xliff=False, parallel=False, no_interactive=False,
            use_git_timestamps=False
        )
        pr_instance.pull.assert_has_calls([pull_call])

    @patch('txclib.utils.get_current_branch')
    @patch('txclib.commands.logger')
    @patch('txclib.commands.project.Project')
    def test_pull_with_branch_and_branchname_option(
        self, project_mock, log_mock, bmock
    ):
        pr_instance = MagicMock()
        project_mock.return_value = pr_instance
        bmock.return_value = None
        cmd_pull(['--branch', 'somebranch'], '.')
        self.assertEqual(pr_instance.pull.call_count, 1)
        pull_call = call(
            branch='somebranch', fetchall=False, fetchsource=False,
            force=False, languages=[], minimum_perc=None, mode=None,
            overwrite=True, pseudo=False, resources=[], skip=False,
            xliff=False, parallel=False, no_interactive=False,
            use_git_timestamps=False
        )
        pr_instance.pull.assert_has_calls([pull_call])

    @patch('txclib.commands.project')
    def test_pull_with_no_interactive(self, project_mock):
        pr_instance = MagicMock()
        pr_instance.pull.return_value = True
        project_mock.Project.return_value = pr_instance
        cmd_pull(['--no-interactive'], '.')
        pull_call = call(
            fetchall=False, force=False, minimum_perc=None,
            skip=False, no_interactive=True, resources=[], pseudo=False,
            languages=[], fetchsource=False, mode=None, branch=None,
            xliff=False, parallel=False, overwrite=True,
            use_git_timestamps=False
        )
        self.assertEqual(pr_instance.pull.call_count, 1)
        pr_instance.pull.assert_has_calls([pull_call])

    @patch('txclib.commands.project')
    def test_pull_with_git_timestamps(self, project_mock):
        pr_instance = MagicMock()
        pr_instance.pull.return_value = True
        project_mock.Project.return_value = pr_instance
        cmd_pull(['--use-git-timestamps'], '.')
        pull_call = call(
            fetchall=False, force=False, minimum_perc=None,
            skip=False, no_interactive=False, resources=[], pseudo=False,
            languages=[], fetchsource=False, mode=None, branch=None,
            xliff=False, parallel=False, overwrite=True,
            use_git_timestamps=True
        )
        self.assertEqual(pr_instance.pull.call_count, 1)
        pr_instance.pull.assert_has_calls([pull_call])


class TestConfigCommand(unittest.TestCase):

    def setUp(self):
        self.curr_dir = os.getcwd()
        os.chdir('./tests/project_dir/')
        os.mkdir('.tx')
        self.path_to_tx = os.getcwd()
        self.config_file = '.tx/config'
        self.config_fd = open(self.config_file, "w")
        self.config_fd.write("[main]\nhost = https://foo.var\n")
        self.config_fd.close()

    def tearDown(self, *args, **kwargs):
        shutil.rmtree('.tx', ignore_errors=False, onerror=None)
        os.chdir(self.curr_dir)
        super(TestConfigCommand, self).tearDown(*args, **kwargs)

    def test_bare_set_too_few_arguments(self):
        with self.assertRaises(SystemExit):
            args = ["-r", "project1.resource1"]
            cmd_config(args, None)

    def test_bare_set_source_no_file(self):
        with self.assertRaises(SystemExit):
            args = ["-r", "project1.resource1", '--is-source', '-l', 'en']
            cmd_config(args, None)

        with self.assertRaises(Exception):
            args = ['-r', 'project1.resource1', '--source', '-l', 'en',
                    'noexistent-file.txt']
            cmd_config(args, self.path_to_tx)

    def test_bare_set_source_file(self):
        expected = ("[main]\nhost = https://foo.var\n\n[project1.resource1]\n"
                    "source_file = test.txt\nsource_lang = en\n\n")
        args = ["-r", "project1.resource1", '--source', '-l', 'en', 'test.txt']
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

        # set translation file for de
        expected = ("[main]\nhost = https://foo.var\n\n[project1.resource1]\n"
                    "source_file = test.txt\nsource_lang = en\n"
                    "trans.de = translations/de.txt\n\n")
        args = ["-r", "project1.resource1", '-l', 'de', 'translations/de.txt']
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_auto_locale_no_expression(self):
        error_msg = "You need to specify an expression"
        with assertRaisesRegex(self, Exception, error_msg):
            args = [MAPPING, "-r", "project1.resource1",
                    '--source-language', 'en']
            cmd_config(args, self.path_to_tx)

    def test_auto_locale(self):
        expected = "[main]\nhost = https://foo.var\n"
        args = [MAPPING, "-r", "project1.resource1", '--source-language',
                'en', 'translations/<lang>/test.txt']
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_auto_locale_is_backwards_compatible(self):
        expected = ("[main]\nhost = https://foo.var\n\n[project1.resource1]\n"
                    "file_filter = translations/<lang>/test.txt\n"
                    "source_file = translations/en/test.txt\n"
                    "source_lang = en\n\n")

        args = ["--auto-local", "-r", "project1.resource1",
                '--source-language', 'en', '--execute',
                'translations/<lang>/test.txt']
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_auto_locale_execute(self):
        expected = ("[main]\nhost = https://foo.var\n\n[project1.resource1]\n"
                    "file_filter = translations/<lang>/test.txt\n"
                    "source_file = translations/en/test.txt\n"
                    "source_lang = en\n\n")

        args = [MAPPING, "-r", "project1.resource1", '--source-language',
                'en', '--execute', 'translations/<lang>/test.txt']
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_auto_remote_invalid_url(self):
        # no project_url
        args = [MAPPINGREMOTE]
        with self.assertRaises(SystemExit):
            cmd_config(args, self.path_to_tx)

        # invalid project_url
        args = [MAPPINGREMOTE, "http://some.random.url/"]
        with self.assertRaises(Exception):
            cmd_config(args, self.path_to_tx)

    @patch('txclib.utils.get_details')
    @patch('txclib.project.Project._extension_for')
    def test_auto_remote_project(self, extension_mock, get_details_mock):
        # change the host to tx
        open(self.config_file, "w").write(
            '[main]\nhost = https://www.transifex.com\n'
        )
        expected = ("[main]\nhost = https://www.transifex.com\n\n"
                    "[proj.resource_1]\n"
                    "file_filter = translations/proj.resource_1/<lang>.txt\n"
                    "source_lang = fr\ntype = TXT\n\n[proj.resource_2]\n"
                    "file_filter = translations/proj.resource_2/<lang>.txt\n"
                    "source_lang = fr\ntype = TXT\n\n")
        extension_mock.return_value = ".txt"
        get_details_mock.side_effect = [
            # project details
            {
                'resources': [
                    {'slug': 'resource_1', 'name': 'resource 1'},
                    {'slug': 'resource_2', 'name': 'resource 2'}
                ]
            },
            # resources details
            {
                'source_language_code': 'fr',
                'i18n_type': 'TXT',
                'slug': 'resource_1',
            }, {
                'source_language_code': 'fr',
                'i18n_type': 'TXT',
                'slug': 'resource_2',
            }
        ]
        args = [MAPPINGREMOTE, "https://www.transifex.com/test-org/proj/"]
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    @patch('txclib.utils.get_details')
    @patch('txclib.project.Project._extension_for')
    def test_auto_remote_is_backwards_compatible(self, extension_mock,
                                                 get_details_mock):
        # change the host to tx
        open(self.config_file, "w").write(
            '[main]\nhost = https://www.transifex.com\n'
        )
        expected = ("[main]\nhost = https://www.transifex.com\n\n"
                    "[proj.resource_1]\n"
                    "file_filter = translations/proj.resource_1/<lang>.txt\n"
                    "source_lang = fr\ntype = TXT\n\n[proj.resource_2]\n"
                    "file_filter = translations/proj.resource_2/<lang>.txt\n"
                    "source_lang = fr\ntype = TXT\n\n")
        extension_mock.return_value = ".txt"
        get_details_mock.side_effect = [
            # project details
            {
                'resources': [
                    {'slug': 'resource_1', 'name': 'resource 1'},
                    {'slug': 'resource_2', 'name': 'resource 2'}
                ]
            },
            # resources details
            {
                'source_language_code': 'fr',
                'i18n_type': 'TXT',
                'slug': 'resource_1',
            }, {
                'source_language_code': 'fr',
                'i18n_type': 'TXT',
                'slug': 'resource_2',
            }
        ]
        args = ["--auto-remote", "https://www.transifex.com/test-org/proj/"]
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_bulk_missing_options(self):
        with self.assertRaises(SystemExit):
            args = [MAPPINGBULK]
            cmd_config(args, self.path_to_tx)

        with self.assertRaises(SystemExit):
            args = [MAPPINGBULK, "-p", "test-project"]
            cmd_config(args, self.path_to_tx)

        with self.assertRaises(SystemExit):
            args = [MAPPINGBULK, "-p", "test-project", "--source-language",
                    "en", "--t", "TXT", "--file-extension", ".txt"]
            cmd_config(args, self.path_to_tx)

    def test_bulk(self):
        expected = ("[main]\nhost = https://foo.var\n\n"
                    "[test-project.translations_en_test]\n"
                    "file_filter = translations/<lang>/en/test.txt\n"
                    "source_file = translations/en/test.txt\n"
                    "source_lang = en\ntype = TXT\n\n")
        args = [MAPPINGBULK, "-p", "test-project", "--source-file-dir",
                "translations", "--source-language", "en", "-t", "TXT",
                "--file-extension", ".txt", "--execute", "--expression",
                "translations/<lang>/{filepath}/{filename}{extension}"]
        cmd_config(args, self.path_to_tx)
        with open(self.config_file) as config:
            self.assertEqual(config.read(), expected)

    def test_invalid_expression(self):
        args = [MAPPINGBULK, "-p", "test-project", "--source-file-dir",
                "translations", "--source-language", "en", "-t", "TXT",
                "--file-extension", ".txt", "--execute", "--expression",
                "expression/{filename}{extension}"]
        with self.assertRaises(SystemExit):
            cmd_config(args, self.path_to_tx)

    def test_invalid_expression_status(self):
        """
        Test how the client handles invalid expressions in file_filter
        """
        valid_expression = "test_expressions/<lang>/test.txt"
        invalid_expression = "test_expressions/{filename}{ext}"
        args = [MAPPINGBULK, "-p", "test-project", "--source-file-dir",
                "test_expressions/en", "--source-language", "en", "-t", "TXT",
                "--file-extension", ".txt", "--execute", "--expression",
                valid_expression]
        # Configure the client using a valid expression
        cmd_config(args, self.path_to_tx)

        # A trick to capture the standard output
        sys_stdout = sys.stdout
        out = StringIO()
        sys.stdout = out
        # Check for the expected status using the valid expression
        cmd_status([], None)
        output = out.getvalue().strip()
        sys.stdout = sys_stdout
        self.assertTrue("test_expressions/en/test.txt" in output)
        self.assertTrue("test_expressions/es/test.txt" in output)

        # Change the configuration file "by hand" and replace the file_filter
        # with an invalid expression
        for line in fileinput.input(self.config_file, inplace=True):
            print(line.replace(valid_expression, invalid_expression))

        sys_stdout = sys.stdout
        out = StringIO()
        sys.stdout = out
        cmd_status([], None)
        output = out.getvalue().strip()
        sys.stdout = sys_stdout
        # Check that the Spanish translation file in not tracked
        self.assertTrue("test_expressions/en/test.txt" in output)
        self.assertFalse("test_expressions/es/test.txt" in output)


class TestMainCommand(unittest.TestCase):
    def test_call_tx_with_no_command(self):
        with self.assertRaises(SystemExit):
            main(['tx'])

    def test_call_tx_with_invalid_command(self):
        with self.assertRaises(SystemExit):
            main(['tx', 'random'])



# -*- coding: utf-8 -*-
"""
In this file we have all the top level commands for the Transifex client. Since
we're using a way to automatically list them and execute them, when adding code
to this file you must take care of the following:
 * Added functions must begin with 'cmd_' followed by the actual name of the
   command being used in the command line (e.g. cmd_init)
 * The description for each function that we display to the user is read from
   the func_doc attribute which reads the doc string. So, when adding
   docstring to a new function make sure you add an one-liner which is
   descriptive and is meant to be seen by the user.
 * When including libraries, it's best if you include modules instead of
   functions because that way our function resolution will work faster and the
   chances of overlapping are minimal
 * All functions should use the OptionParser and should have a usage and
   description field.
"""
import os
import sys
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import Namespace

from txclib import utils, project
from txclib.config import OrderedRawConfigParser
from txclib.exceptions import UnInitializedError
from txclib.parsers import delete_parser, help_parser, parse_csv_option, \
    status_parser, pull_parser, set_parser, push_parser, init_parser, \
    MAPPING, MAPPINGREMOTE, MAPPINGBULK
from txclib.paths import posix_path
from txclib.log import logger
from txclib.wizard import Wizard
from txclib import messages


def cmd_init(argv, path_to_tx):
    """Initialize a new Transifex project."""
    parser = init_parser()
    options = parser.parse_args(argv)
    path_to_tx = options.path_to_tx or os.getcwd()

    print(messages.init_intro)
    save = options.save
    # if we already have a config file and we are not told to override it
    # in the args we have to ask
    config_file = os.path.join(path_to_tx, ".tx", "config")
    if os.path.isfile(config_file):
        if not save:
            if options.no_interactive:
                parser.error("Project already initialized.")
            logger.info(messages.init_initialized)
            if not utils.confirm(messages.init_reinit):
                return
        os.remove(config_file)

    if not os.path.isdir(os.path.join(path_to_tx, ".tx")):
        logger.info("Creating .tx folder...")
        os.mkdir(os.path.join(path_to_tx, ".tx"))

    default_transifex = "https://www.transifex.com"
    transifex_host = options.host or default_transifex

    if not transifex_host.startswith(('http://', 'https://')):
        transifex_host = 'https://' + transifex_host

    if not os.path.exists(config_file):
        # Handle the credentials through transifexrc
        config = OrderedRawConfigParser()
        config.add_section('main')
        config.set('main', 'host', transifex_host)
        # Touch the file if it doesn't exist
        logger.info("Creating config file...")
        fh = open(config_file, 'w')
        config.write(fh)
        fh.close()

    if not options.skipsetup and not options.no_interactive:
        logger.info(messages.running_tx_set)
        cmd_config([], path_to_tx)
    else:
        prj = project.Project(path_to_tx)
        prj.getset_host_credentials(transifex_host, username=options.user,
                                    password=options.password,
                                    token=options.token, force=options.save,
                                    no_interactive=options.no_interactive)
        prj.save()
        logger.info("Done.")


def cmd_set(argv, path_to_tx):
    """
    Add local or remote files under Transifex. Warning: This command will be
    deprecated in a future release of the client. You should use the
    `tx config` command.
    """
    cmd_config(argv, path_to_tx, is_legacy=True)


def cmd_config(argv, path_to_tx, is_legacy=False):
    """Add local or remote files under Transifex"""
    from_wizard = False
    if len(argv) == 0:
        # since interactive wizard should be equivalent to auto-local
        # subcommand there are some default options that need to be set
        default_options = {
            'execute': True,
            'subcommand': MAPPING,
            'minimum_perc': 0,
            'mode': None,
            'expression_legacy': None,
        }
        try:
            # Run the wizard and configure parse with the wizard inputs
            wizard_options = Wizard(path_to_tx).run()
            wizard_options.update(default_options)
            options = Namespace(**wizard_options)
            parser = set_parser(is_legacy=is_legacy)
            from_wizard = True
        except SystemExit:
            print("\n")
            sys.exit(1)
    else:
        # hack to support backwards compatibility
        # for --auto-local and --auto-remote options
        if '--auto-local' in argv:
            argv.pop(argv.index('--auto-local'))
            argv.insert(0, MAPPING)
        if '--auto-remote' in argv:
            argv.pop(argv.index('--auto-remote'))
            argv.insert(0, MAPPINGREMOTE)

        is_subcommand = argv[0] in CONFIG_SUBCOMMANDS.keys()
        parser = set_parser(subparser=is_subcommand, is_legacy=is_legacy)
        options = parser.parse_args(argv)

    _validate_set_arguments(parser, options)

    if not hasattr(options, 'subcommand'):
        bare_set(path_to_tx, options)
    else:
        try:
            CONFIG_SUBCOMMANDS[options.subcommand](
                path_to_tx, options, from_wizard=from_wizard
            )
        except KeyError:
            parser.print_help()
            sys.exit(2)


def _validate_set_arguments(parser, options):
    """
    Do any extra required validation for set command arguments
    """
    if hasattr(options, 'is_source') and hasattr(options, 'language'):
        if options.is_source and not options.language:
            parser.error("Please specify a source language.")

    if hasattr(options, 'expression'):
        if options.expression and '<lang>' not in options.expression:
            parser.error("The expression you have provided is not valid.")

    if hasattr(options, 'resource') and not\
            utils.valid_resource_slug(options.resource):
        parser.error("Invalid resource slug. The format is <project_slug>"
                     ".<resource_slug> and the valid characters include "
                     "[_-\w].")


def bare_set(path_to_tx, options):
    filename = options.filename
    # Calculate relative path
    path_to_file = os.path.relpath(filename, path_to_tx)

    if options.is_source:
        resource = options.resource
        _set_source_file(path_to_tx, resource, options.language, path_to_file)
    elif options.resource or options.language:
        resource = options.resource
        lang = options.language

        _go_to_dir(path_to_tx)
        _set_translation(path_to_tx, resource, lang, path_to_file)

    _set_mode(options.resource, options.mode, path_to_tx)
    _set_type(options.resource, options.i18n_type, path_to_tx)
    _set_minimum_perc(options.resource, options.minimum_perc, path_to_tx)

    logger.info("Done.")


def subcommand_mapping(path_to_tx, options, from_wizard=False):
    expression = options.expression or options.expression_legacy
    _auto_local(path_to_tx, options.resource,
                source_language=options.source_language,
                expression=expression, source_file=options.source_file,
                execute=options.execute, regex=False)
    if options.execute:
        _set_minimum_perc(options.resource, options.minimum_perc,
                          path_to_tx)
        _set_mode(options.resource, options.mode, path_to_tx)
        _set_type(options.resource, options.i18n_type, path_to_tx)

    if from_wizard:
        _print_instructions(options.resource, path_to_tx)


def subcommand_mapping_remote(path_to_tx, options, **kwargs):
    url = options.project_url
    _auto_remote(path_to_tx, url)
    _set_minimum_perc(None, options.minimum_perc, path_to_tx)
    _set_mode(None, options.mode, path_to_tx)


def _print_instructions(resource, path_to_tx):
    keys = ['source_file', 'file_filter', 'source_lang', 'type']
    prj = project.Project(path_to_tx)
    fmt_kwargs = {k: prj.config.get(resource, k) for k in keys}
    fmt_kwargs.update({'resource': resource})
    print(messages.final_instr.format(**fmt_kwargs))


def subcommand_mapping_bulk(path_to_tx, options, **kwargs):
    """Add local files for multiple resources under Transifex"""

    if not options.file_extension.startswith('.'):
        file_extension = '.{}'.format(options.file_extension)
    else:
        file_extension = options.file_extension

    curpath = os.path.join(path_to_tx, options.source_file_dir)

    for root, dirs, files in os.walk(curpath):
        for file_name in files:
            if not file_name.endswith(file_extension):
                continue

            source_file = os.path.join(
                root, file_name
            ).split(path_to_tx)[1][1:]
            resource = source_file.split('.')[0].replace('/', '_')
            relpath = root.split(options.source_file_dir)[1]

            should_continue = True
            for directory in options.ignore_dirs:
                if relpath.strip('/').startswith(directory):
                    should_continue = False
                    break

            if not should_continue:
                break

            expr = options.expression.format(
                filepath=relpath, filename=file_name.split('.')[0],
                extension=options.file_extension
            ).replace('//', '/')

            full_resource = '{}.{}'.format(options.project, resource)
            _auto_local(
                path_to_tx, full_resource, options.source_language, expr,
                source_file=source_file, execute=options.execute, regex=False,
            )

            if options.execute:
                _set_minimum_perc(full_resource, options.minimum_perc,
                                  path_to_tx)
                _set_mode(full_resource, options.mode, path_to_tx)
                _set_type(full_resource, options.i18n_type, path_to_tx)


def _auto_local(path_to_tx, resource, source_language, expression,
                execute=False, source_file=None, regex=False):
    """Auto configure local project."""
    # The path everything will be relative to
    curpath = os.path.abspath(os.curdir)

    # Force expr to be a valid regex expr (escaped) but keep <lang> intact
    if not expression:
        raise Exception("You need to specify an expression to define where "
                        "translation files should be saved.")

    if not execute:
        logger.info("Only printing the commands which will be run if the "
                    "--execute switch is specified.")

    # First, let's construct a dictionary of all matching files.
    # Note: Only the last matching file of a language will be stored.
    translation_files = {}
    for f_path, lang in utils.get_project_files(curpath, expression):
        if lang == source_language and not source_file:
            source_file = f_path
        else:
            translation_files[lang] = f_path

    if not source_file:
        raise Exception("Could not find a source language file. Please run "
                        "set --source manually and then re-run this command "
                        "or provide the source file with the -s flag.")
    if execute:
        logger.info("Updating source for resource %s ( %s -> %s )." % (
                    resource, source_language, os.path.relpath(
                        source_file, path_to_tx)
                    ))
        _set_source_file(path_to_tx, resource, source_language,
                         os.path.relpath(source_file, path_to_tx))
    else:
        logger.info('\ntx set --source -r %(res)s -l %(lang)s %(file)s\n' % {
            'res': resource,
            'lang': source_language,
            'file': os.path.relpath(source_file, curpath)})

    prj = project.Project(path_to_tx)

    if execute:
        try:
            prj.config.get("%s" % resource, "source_file")
        except configparser.NoSectionError:
            raise Exception("No resource with slug \"%s\" was found.\nRun "
                            "'tx set auto-local -r %s \"expression\"' to "
                            "do the initial configuration." % resource)

    # Now let's handle the translation files.
    if execute:
        logger.info("Updating file expression for resource %s ( %s )." % (
                    resource, expression))
        # Evaluate file_filter relative to root dir
        file_filter = posix_path(
            os.path.relpath(os.path.join(curpath, expression), path_to_tx)
        )
        prj.config.set("%s" % resource, "file_filter", file_filter)
    else:
        for (lang, f_path) in sorted(translation_files.items()):
            logger.info('tx set -r %(res)s -l %(lang)s %(file)s' % {
                'res': resource,
                'lang': lang,
                'file': os.path.relpath(f_path, curpath)})

    if execute:
        prj.save()


def _auto_remote(path_to_tx, url):
    """Initialize a remote project/resource to the current directory."""
    logger.info("Auto configuring local project from remote URL...")

    type, vars = utils.parse_tx_url(url)
    prj = project.Project(path_to_tx)
    username, password = prj.getset_host_credentials(vars['hostname'])

    if type.startswith('project'):
        logger.info("Getting details for project %s" % vars['project'])
        proj_info = utils.get_details(
            'project_details',
            username, password,
            hostname=vars['hostname'],
            project=vars['project'])
        resources = ['.'.join([vars['project'],
                     r['slug']]) for r in proj_info['resources']]
        logger.info("%s resources found. Configuring..." % len(resources))
    elif type == 'release':
        logger.info("Getting details for release %s" % vars['release'])
        rel_info = utils.get_details(
            'release_details',
            username, password,
            hostname=vars['hostname'],
            project=vars['project'],
            release=vars['release'])
        resources = []
        for r in rel_info['resources']:
            if 'project' in r:
                resources.append('.'.join([r['project']['slug'], r['slug']]))
            else:
                resources.append('.'.join([vars['project'], r['slug']]))
        logger.info("%s resources found. Configuring..." % len(resources))
    elif type.startswith('resource'):
        logger.info("Getting details for resource %s" % vars['resource'])
        resources = ['.'.join([vars['project'], vars['resource']])]
    else:
        raise Exception("Url '%s' is not recognized." % url)

    for resource in resources:
        logger.info("Configuring resource %s." % resource)
        proj, res = resource.split('.')
        res_info = utils.get_details(
            'resource_details',
            username, password,
            hostname=vars['hostname'],
            project=proj, resource=res)
        try:
            source_lang = res_info['source_language_code']
            i18n_type = res_info['i18n_type']
        except KeyError:
            raise Exception("Remote server seems to be running an unsupported "
                            "version of Transifex. Either update your server "
                            "software of fallback to a previous version "
                            "of transifex-client.")
        prj.set_remote_resource(
            resource=resource,
            host=vars['hostname'],
            source_lang=source_lang,
            i18n_type=i18n_type)

    prj.save()


def cmd_push(argv, path_to_tx):
    """Push local files to remote server"""
    parser = push_parser()
    options = parser.parse_args(argv)
    force_creation = options.force_creation
    languages = parse_csv_option(options.languages)
    resources = parse_csv_option(options.resources)
    skip = options.skip_errors
    xliff = options.xliff
    parallel = options.parallel
    use_git_timestamps = options.use_git_timestamps
    prj = project.Project(path_to_tx)
    if not (options.push_source or options.push_translations):
        parser.error("You need to specify at least one of the -s|--source, "
                     "-t|--translations flags with the push command.")

    branch = get_branch_from_options(options, prj.root)
    prj.push(
        force=force_creation, resources=resources, languages=languages,
        skip=skip, source=options.push_source,
        translations=options.push_translations,
        no_interactive=options.no_interactive,
        xliff=xliff, branch=branch, parallel=parallel,
        use_git_timestamps=use_git_timestamps,
    )
    logger.info("Done.")


def cmd_pull(argv, path_to_tx):
    """Pull files from remote server to local repository"""
    parser = pull_parser()
    options = parser.parse_args(argv)
    if options.fetchall and options.languages:
        parser.error("You can't user a language filter along with the "
                     "-a|--all option")
    languages = parse_csv_option(options.languages)
    resources = parse_csv_option(options.resources)
    pseudo = options.pseudo
    # Should we download as xliff?
    xliff = options.xliff
    parallel = options.parallel
    skip = options.skip_errors
    minimum_perc = options.minimum_perc or None
    use_git_timestamps = options.use_git_timestamps

    _go_to_dir(path_to_tx)

    # instantiate the project.Project
    prj = project.Project(path_to_tx)
    branch = get_branch_from_options(options, prj.root)
    prj.pull(
        languages=languages, resources=resources, overwrite=options.overwrite,
        fetchall=options.fetchall, fetchsource=options.fetchsource,
        force=options.force, skip=skip, minimum_perc=minimum_perc,
        mode=options.mode, pseudo=pseudo, xliff=xliff, branch=branch,
        parallel=parallel, no_interactive=options.no_interactive,
        use_git_timestamps=use_git_timestamps
    )
    logger.info("Done.")


def _set_source_file(path_to_tx, resource, lang, path_to_file):
    """Reusable method to set source file."""
    proj, res = resource.split('.')
    if not proj or not res:
        raise Exception("\"%s.%s\" is not a valid resource identifier. "
                        "It should be in the following format "
                        "project_slug.resource_slug." %
                        (proj, res))
    if not lang:
        raise Exception("You haven't specified a source language.")

    _go_to_dir(path_to_tx)

    if not os.path.exists(path_to_file):
        raise Exception("tx: File ( %s ) does not exist." %
                        os.path.join(path_to_tx, path_to_file))

    # instantiate the project.Project
    prj = project.Project(path_to_tx)
    root_dir = os.path.abspath(path_to_tx)

    if root_dir not in os.path.normpath(os.path.abspath(path_to_file)):
        raise Exception("File must be under the project root directory.")

    logger.info("Setting source file for resource %s.%s ( %s -> %s )." % (
        proj, res, lang, path_to_file))

    path_to_file = os.path.relpath(path_to_file, root_dir)

    prj = project.Project(path_to_tx)

    # FIXME: Check also if the path to source file already exists.
    try:
        try:
            prj.config.get("%s.%s" % (proj, res), "source_file")
        except configparser.NoSectionError:
            prj.config.add_section("%s.%s" % (proj, res))
        except configparser.NoOptionError:
            pass
    finally:
        prj.config.set(
            "%s.%s" % (proj, res), "source_file", posix_path(path_to_file)
        )
        prj.config.set("%s.%s" % (proj, res), "source_lang", lang)

    prj.save()


def _set_translation(path_to_tx, resource, lang, path_to_file):
    """Reusable method to set translation file."""

    proj, res = resource.split('.')
    if not project or not resource:
        raise Exception("\"%s\" is not a valid resource identifier. "
                        "It should be in the following format "
                        "project_slug.resource_slug." %
                        resource)

    _go_to_dir(path_to_tx)

    # Warn the user if the file doesn't exist
    if not os.path.exists(path_to_file):
        logger.info("Warning: File '%s' doesn't exist." % path_to_file)

    # instantiate the project.Project
    prj = project.Project(path_to_tx)
    root_dir = os.path.abspath(path_to_tx)

    if root_dir not in os.path.normpath(os.path.abspath(path_to_file)):
        raise Exception("File must be under the project root directory.")

    if lang == prj.config.get("%s.%s" % (proj, res), "source_lang"):
        raise Exception("tx: You cannot set translation file for "
                        "the source language. Source languages contain "
                        "the strings which will be translated!")

    logger.info("Updating translations for resource %s ( %s -> %s )." % (
                resource, lang, path_to_file))
    path_to_file = os.path.relpath(path_to_file, root_dir)
    prj.config.set(
        "%s.%s" % (proj, res), "trans.%s" % lang, posix_path(path_to_file)
    )

    prj.save()


def cmd_status(argv, path_to_tx):
    """Print status of current project"""
    parser = status_parser()
    options = parser.parse_args(argv)
    resources = parse_csv_option(options.resources)
    prj = project.Project(path_to_tx)
    resources = prj.get_chosen_resources(resources)
    resources_num = len(resources)
    for idx, res in enumerate(resources):
        p, r = res.split('.')
        print("%s -> %s (%s of %s)" % (p, r, idx + 1, resources_num))
        print("Translation Files:")
        slang = prj.get_resource_option(res, 'source_lang')
        sfile = prj.get_resource_option(res, 'source_file') or "N/A"
        lang_map = prj.get_resource_lang_mapping(res)
        print(" - %s: %s (%s)" % (utils.color_text(slang, "RED"),
              sfile, utils.color_text("source", "YELLOW")))
        files = prj.get_resource_files(res)
        fkeys = list(files.keys())
        fkeys.sort()
        for lang in fkeys:
            local_lang = lang
            if lang in list(lang_map.values()):
                local_lang = lang_map.flip[lang]
            print(" - %s: %s" % (utils.color_text(local_lang, "RED"),
                  files[lang]))
        print("")


def cmd_help(argv, path_to_tx):
    """List all available commands"""
    parser = help_parser()
    options = parser.parse_args(argv)

    # Get all commands
    fns = utils.discover_commands()

    # Print help for specific command
    if options.command:
        try:
            fns[options.command](['--help'], path_to_tx)
        except KeyError:
            utils.logger.error("Command %s not found" % argv[0])
    # or print summary of all commands

    # the code below will only be executed if the KeyError exception is thrown
    # because in all other cases the function called with --help will exit
    # instead of return here
    keys = list(fns.keys())
    keys.sort()

    print("Transifex command line client.\n")
    print("Available commands are:")
    for key in keys:
        print("  %-15s\t%s" % (key, getattr(fns[key], '__doc__')))
    print("\nFor more information run %s command --help" % sys.argv[0])


def cmd_delete(argv, path_to_tx):
    """Delete an accessible resource or translation in a remote server."""
    parser = delete_parser()
    options = parser.parse_args(argv)
    languages = parse_csv_option(options.languages)
    resources = parse_csv_option(options.resources)
    skip = options.skip_errors
    force = options.force_delete
    prj = project.Project(path_to_tx)
    prj.delete(resources, languages, skip, force)
    logger.info("Done.")


def _go_to_dir(path):
    """Change the current working directory to the directory specified as
    argument.

    Args:
        path: The path to chdir to.
    Raises:
        UnInitializedError, in case the directory has not been initialized.
    """
    if path is None:
        raise UnInitializedError(
            "Directory has not been initialized. "
            "Did you forget to run 'tx init' first?"
        )
    os.chdir(path)


def _set_minimum_perc(resource, value, path_to_tx):
    """Set the minimum percentage in the .tx/config file."""
    args = (resource, 'minimum_perc', value, path_to_tx, 'set_min_perc')
    _set_project_option(*args)


def _set_mode(resource, value, path_to_tx):
    """Set the mode in the .tx/config file."""
    args = (resource, 'mode', value, path_to_tx, 'set_default_mode')
    _set_project_option(*args)


def _set_type(resource, value, path_to_tx):
    """Set the i18n type in the .tx/config file."""
    args = (resource, 'type', value, path_to_tx, 'set_i18n_type')
    _set_project_option(*args)


def _set_project_option(resource, name, value, path_to_tx, func_name):
    """Save the option to the project configuration file."""
    if value is None:
        return
    if not resource:
        logger.debug("Setting the %s for all resources." % name)
        resources = []
    else:
        logger.debug("Setting the %s for resource %s." % (name, resource))
        resources = [resource, ]
    prj = project.Project(path_to_tx)
    getattr(prj, func_name)(resources, value)
    prj.save()


def get_branch_from_options(options, project_root):
    """ Returns the branch name that needs to be used in command
    based on parser options.

    options: optparse parser options as returned from `parse()`
    project_root: project root directory
    """
    if not options.branch:
        return

    if options.branch != '-1':
        return options.branch

    branch = utils.get_current_branch(project_root)
    if not branch:
        logger.error("You specified the --branch option but current "
                     "directory does not seem to belong in any git repo.")
        sys.exit(1)
    return branch


CONFIG_SUBCOMMANDS = {
    MAPPING: subcommand_mapping,
    MAPPINGREMOTE: subcommand_mapping_remote,
    MAPPINGBULK: subcommand_mapping_bulk,
}



# -*- coding: utf-8 -*-

import os
import platform
from pkg_resources import resource_filename
import txclib


cacerts_file = None

def user_agent_identifier():
    """Return the user agent for the client."""
    client_info = (txclib.__version__, platform.system(), platform.machine())
    return "txclient/%s (%s %s)" % client_info


def certs_file():
    return cacerts_file or system_certs_file()


def system_certs_file():
    if platform.system() == 'Windows':
        return os.path.join(txclib.utils.get_base_dir(), 'cacert.pem')
    else:
        POSSIBLE_CA_BUNDLE_PATHS = [
            # Red Hat, CentOS, Fedora and friends
            # (provided by the ca-certificates package):
            '/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem',
            '/etc/ssl/certs/ca-bundle.crt',
            '/etc/pki/tls/certs/ca-bundle.crt',
            # Ubuntu, Debian, and friends
            # (provided by the ca-certificates package):
            '/etc/ssl/certs/ca-certificates.crt',
            # FreeBSD (provided by the ca_root_nss package):
            '/usr/local/share/certs/ca-root-nss.crt',
            # openSUSE (provided by the ca-certificates package),
            # the 'certs' directory is the
            # preferred way but may not be supported by the SSL module,
            # thus it has 'ca-bundle.pem'
            # as a fallback (which is generated from pem files in the
            # 'certs' directory):
            '/etc/ssl/ca-bundle.pem',
        ]
        for path in POSSIBLE_CA_BUNDLE_PATHS:
            if os.path.exists(path):
                return path
        return resource_filename(__name__, 'cacert.pem')



# -*- coding: utf-8 -*-

import os
import fnmatch
import datetime
import time
import urllib3
import six

try:
    import urlparse
    from urllib import urlencode
except ImportError:  # For Python 3
    import urllib.parse as urlparse
    from urllib.parse import urlencode
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

from requests.exceptions import HTTPError
from slugify import slugify

from txclib import web
from txclib import api
from txclib import utils
from txclib import messages
from urllib3.exceptions import SSLError
from six.moves import input
from txclib.exceptions import (
    ConfigFileError, HttpNotFound, HttpNotAuthorized, MalformedConfigFile,
    TransifexrcConfigFileError
)
from txclib.urls import API_URLS
from txclib.config import Flipdict, CERT_REQUIRED
from txclib.log import logger
from txclib.paths import native_path, posix_sep
from txclib.utils import ProjectNotInit, perform_parallel_requests


DEFAULT_PULL_URL = 'pull_file'
PULL_MODE_URL_NAME = "pull_{mode}_file"

DEFAULT_API_HOSTNAME = "https://api.transifex.com"
API_USERNAME = api.Api.USERNAME


class Project(object):
    """
    Represent an association between the local and remote project instances.
    """

    SKIP_DECODE_I18N_TYPES = ['DOCX', 'XLSX']
    FILE_FILTER = "translations<sep>%(proj)s.%(res)s<sep><lang>.%(extension)s"

    def __init__(self, path_to_tx=None, init=True):
        """Initialize the Project attributes."""
        if init:
            self._init(path_to_tx)

    def _init(self, path_to_tx=None):
        instructions = "Run 'tx init' to initialize your project first!"
        try:
            self.root = utils.get_tx_dir_path(path_to_tx)
            self.config_file = utils.get_config_file_path(self.root)
            self.config = utils.read_config_file(self.config_file)

            self.txrc_file = utils.get_transifex_file()
            self.txrc = utils.get_transifex_config(self.txrc_file)
        except ProjectNotInit:
            logger.error(instructions)
            raise
        host = self.config.get('main', 'host')
        if host.lower().startswith('https://'):
            self.conn = urllib3.connection_from_url(
                host,
                cert_reqs=CERT_REQUIRED,
                ca_certs=web.certs_file()
            )
        else:
            self.conn = urllib3.connection_from_url(host)

    def validate_config(self):
        """To ensure the json structure is correctly formed."""
        pass

    def validate_credentials(self, username, password, host=None):
        """Check if api credentials are valid."""
        try:
            api.Api(
                username=username, password=password,
                path_to_tx=self.txrc_file, host=host
            ).get('auth_check')
            return True
        except HTTPError as e:
            if e.response.status_code == 401:
                return False
            raise
        return True

    def getset_host_credentials(self, host, username=None, password=None,
                                token=None, no_interactive=False,
                                only_token=False, force=False):
        """Read .transifexrc and report user,
        pass or a token for a specific host else ask the user for input.
        If the credentials provided are different from the .transifexrc file
        ask for confirmation and update them
        """

        save = False
        config_username, config_password = None, None
        if 'TX_TOKEN' in os.environ:
            # User api info from environment variable
            password = os.environ['TX_TOKEN']
            username = API_USERNAME
            """
            We need to check if hostname info exists in the .transifexrc file
            If not, it means that this is the first time this function runs, so
            we need to create its data.
            """
            save = self._add_host_to_config_file(host)
            if save:
                logger.info("Using TX_TOKEN environment variable. Credentials "
                            "won't be saved to .transifexrc.")
                self.save()
            if token:
                # Both env and argument given, inform the user
                logger.warning(
                    "There is a token in the TX_TOKEN env variable and the "
                    "--token argument. The latter will be ignored and no "
                    "credentials will be saved to the .transifexrc file."
                )
            return username, password

        try:
            config_username = self.txrc.get(host, 'username')
            config_password = self.txrc.get(host, 'password')
        except (configparser.NoOptionError, configparser.NoSectionError):
            save = True

        if token:
            password = token
            username = API_USERNAME

        if not (username and password) and not \
               (config_username and config_password):
            username = API_USERNAME
            if not no_interactive:
                password = self._token_prompt(host)
            save = True
        elif config_username and config_password:
            if username == config_username and password == config_password:
                pass
            elif username and password:
                if force:
                    save = True
                elif not no_interactive and utils.confirm(messages.update_txrc):  # noqa
                    save = True
            else:
                username = config_username
                password = config_password

        # In the cases where we need to use the credentials with the new
        # api we can only use a token and not a password so we do an extra
        # validation and prompt the use for a token if the validation fails
        if only_token and not self.validate_credentials(username, password):
            logger.info("You need a valid API token to proceed")
            username = API_USERNAME
            password = self._token_prompt(host)
            save = True

        if save:
            self._add_host_to_config_file(host)
            self.txrc.set(host, 'username', username)
            self.txrc.set(host, 'password', password)
            self.save()
        return username, password

    def _add_host_to_config_file(self, host):
        """Check if a given host exists in .transifexrc, and if not it
        updates the file with the appropriate info.

        :param str host: The hostname to search for and add if needed.
        :return: True if adding a new host, nothing otherwise.
        :rtype: bool
        """
        if not self.txrc.has_section(host):
            logger.info("Updating %s file..." % self.txrc_file)
            self.txrc.add_section(host)
            self.txrc.set(host, 'hostname', host)
            self.txrc.set(host, 'api_hostname',
                          utils.DEFAULT_HOSTNAMES['api_hostname'])
            return True
        return False

    def _token_prompt(self, host):
        token = None
        COLOR = "CYAN"
        while not token:
            token = input(utils.color_text(messages.token_msg, COLOR))
            print(utils.color_text("Verifying token...", "YELLOW"))
            if not self.validate_credentials(API_USERNAME, token, host):
                logger.info(messages.token_validation_failed)
                token = None
        return token

    def set_remote_resource(self, resource, source_lang, i18n_type, host,
                            file_filter=None):
        """Method to handle the addition/configuration of a remote resource."""
        if file_filter is None:
            file_filter = self.FILE_FILTER

        if not self.config.has_section(resource):
            self.config.add_section(resource)

        p_slug, r_slug = resource.split('.', 1)
        file_filter = file_filter.replace("<sep>", r"%s" % posix_sep)
        self._set_url_info(host=host, project=p_slug, resource=r_slug)
        extension = self._extension_for(i18n_type)[1:]

        self.config.set(resource, 'source_lang', source_lang)
        self.config.set(
            resource, 'file_filter',
            file_filter % {'proj': p_slug,
                           'res': r_slug,
                           'extension': extension}
        )
        self.config.set(resource, 'type', i18n_type)
        if host != self.config.get('main', 'host'):
            self.config.set(resource, 'host', host)

    def get_resource_host(self, resource):
        """Return the host that the resource is configured to use.
        If there is no such option we return the default one.
        """
        return self.config.get('main', 'host')

    def get_resource_lang_mapping(self, resource):
        """Get language mappings for a specific resource."""
        lang_map = Flipdict()
        try:
            args = self.config.get("main", "lang_map")
            for arg in args.replace(' ', '').split(','):
                k, v = arg.split(":")
                lang_map.update({k: v})
        except configparser.NoOptionError:
            pass
        except ValueError:
            raise MalformedConfigFile(
                "Your lang map configuration is not correct.")

        res_lang_map = Flipdict()
        if self.config.has_section(resource):
            try:
                args = self.config.get(resource, "lang_map")
                for arg in args.replace(' ', '').split(','):
                    k, v = arg.split(":")
                    res_lang_map.update({k: v})
            except configparser.NoOptionError:
                pass
            except ValueError:
                raise MalformedConfigFile(
                    "Your lang map configuration is not correct.")

        # merge the lang maps and return result
        lang_map.update(res_lang_map)

        return lang_map

    def get_source_file(self, resource):
        """Get source file for a resource."""
        if self.config.has_section(resource):
            source_lang = self.config.get(resource, "source_lang")
            source_file = self.get_resource_option(resource, 'source_file')\
                or None

            if source_file is None:
                try:
                    file_filter = self.config.get(resource, "file_filter")
                    filename = file_filter.replace('<lang>', source_lang)
                    if os.path.exists(filename):
                        return native_path(filename)
                except configparser.NoOptionError:
                    pass
            else:
                return native_path(source_file)

    def get_resource_files(self, resource, xliff=False):
        """Get a dict for all files assigned to a resource.
        First we calculate the files matching the file expression and
        then we apply all translation exceptions.
        The resulting dict will be in this format:

        { 'en': 'path/foo/en/bar.po',
        'de': 'path/foo/de/bar.po',
        'es': 'path/exceptions/es.po'}

        NOTE: All paths are relative to the root of the project
        """
        tr_files = {}
        if self.config.has_section(resource):
            try:
                file_filter = self.config.get(resource, "file_filter")
            except configparser.NoOptionError:
                file_filter = "$^"
            if xliff:
                # update the file-path in case of xliff option
                file_filter += '.xlf'
            source_lang = self.config.get(resource, "source_lang")
            source_file = self.get_source_file(resource)
            matched_files = utils.get_project_files(self.root, file_filter)
            for f_path, lang in matched_files:
                if lang != source_lang:
                    f_path = os.path.relpath(f_path, self.root)
                    if f_path != source_file:
                        tr_files.update({lang: f_path})

            for (name, value) in self.config.items(resource):
                if name.startswith("trans."):
                    value = native_path(value)
                    lang = name.split('.')[1]
                    # delete language which has same file
                    if value in list(tr_files.values()):
                        keys = []
                        for k, v in six.iteritems(tr_files):
                            if v == value:
                                keys.append(k)
                        if len(keys) == 1:
                            del tr_files[keys[0]]
                        else:
                            raise Exception("Your configuration seems wrong. "
                                            "You have multiple languages "
                                            "pointing to the same file.")
                    # Add language with correct file
                    tr_files.update({lang: value})

            return tr_files

        return None

    def get_resource_option(self, resource, option):
        """Return the requested option for a specific resource

        If there is no such option, we return None
        """

        if self.config.has_section(resource):
            if self.config.has_option(resource, option):
                return self.config.get(resource, option)
        return None

    def get_resource_list(self, project=None):
        """Parse configuration file and return tuples with the following format

        [ (project_slug, resource_slug), (..., ...)]
        """

        resource_list = []
        for r in self.config.sections():
            if r == 'main':
                continue
            p_slug, r_slug = r.split('.', 1)
            if project and p_slug != project:
                continue
            resource_list.append(r)

        return resource_list

    def save(self):
        """
        Store the configuration dictionary in the .tx/config file of the
        project.
        """
        utils.save_tx_config(self.config_file, self.config)
        utils.save_txrc_file(self.txrc_file, self.txrc)

    def get_full_path(self, relpath):
        if relpath[0] == os.path.sep:
            return relpath
        else:
            return os.path.join(self.root, relpath)

    def _get_pseudo_file(self, slang, resource, file_filter):
        pseudo_file = file_filter.replace('<lang>', '%s_pseudo' % slang)
        return native_path(pseudo_file)

    def _slug_with_branch(self, resource_slug, branch):
        """ Return the resource slug prefixed by the branch name.
        """

        return '{branch}--{resource}'.format(branch=slugify(branch),
                                             resource=resource_slug)

    def pull(self, languages=None, resources=None, overwrite=True,
             fetchall=False, fetchsource=False, force=False, skip=False,
             minimum_perc=0, mode=None, pseudo=False, xliff=False, branch=None,
             parallel=False, no_interactive=False, use_git_timestamps=False):
        """Pull all translations file from Transifex server."""
        languages = languages or []
        resources = resources or []

        self.minimum_perc = minimum_perc
        resource_list = self.get_chosen_resources(resources)
        skip_decode = False
        params = {}

        url = self._get_url_by_pull_mode(mode=mode)

        for resource in resource_list:
            logger.debug("Handling resource %s" % resource)
            self.resource = resource
            project_slug, resource_slug = resource.split('.', 1)
            if branch:
                resource_slug = self._slug_with_branch(resource_slug, branch)
            files = self.get_resource_files(resource)
            slang = self.get_resource_option(resource, 'source_lang')
            sfile = self.get_source_file(resource)
            lang_map = self.get_resource_lang_mapping(resource)
            host = self.get_resource_host(resource)
            logger.debug("Language mapping is: %s" % lang_map)
            if mode is None:
                mode = self._get_option(resource, 'mode')
            self._set_url_info(host=host, project=project_slug,
                               resource=resource_slug)
            logger.debug("URL data are: %s" % self.url_info)

            try:
                stats = self._get_stats_for_resource(
                    no_interactive=no_interactive
                )
                details_response, _ = self.do_url_request('resource_details')
            except Exception as e:
                if isinstance(e, HttpNotAuthorized):
                    logger.error("Request is not authorized.")
                    continue
                if isinstance(e, HttpNotFound):
                    msg = "Resource %s doesn't exist on the server."
                    logger.error(msg % resource)
                    continue
                if isinstance(e, SSLError) or not skip:
                    raise
                else:
                    msg = "Skipping resource %s"
                    logger.warn(msg % resource)
                    continue

            details = utils.parse_json(details_response)
            if details['i18n_type'] in self.SKIP_DECODE_I18N_TYPES:
                skip_decode = True
            try:
                file_filter = self.config.get(resource, 'file_filter')
            except configparser.NoOptionError:
                file_filter = None

            # Pull source file
            pull_languages = set([])
            new_translations = set([])

            if pseudo:
                pseudo_file = self._get_pseudo_file(
                    slang, resource, file_filter
                )
                if self._should_download(
                        slang, stats, local_file=pseudo_file,
                        use_git_timestamps=use_git_timestamps):
                    logger.info("Pulling pseudo file for resource %s (%s)." % (
                        resource,
                        utils.color_text(pseudo_file, "RED")
                    ))
                    self._download_pseudo(
                        project_slug, resource_slug, pseudo_file
                    )
                if not languages:
                    continue

            if fetchall:
                new_translations = self._new_translations_to_add(
                    files, slang, lang_map, stats, force, use_git_timestamps
                )
                if new_translations:
                    msg = ("New translations found "
                           "for the following languages:%s")
                    logger.info(msg % ', '.join(new_translations))

            existing, new = self._languages_to_pull(
                languages, files, lang_map, stats, force
            )
            pull_languages |= existing
            new_translations |= new
            logger.debug("Adding to new translations: %s" % new)

            if fetchsource:
                if sfile and slang not in pull_languages:
                    pull_languages.add(slang)
                elif slang not in new_translations:
                    new_translations.add(slang)

            if pull_languages:
                prefix = "Queueing" if parallel else "Pulling"
                logger.debug("%s languages for: %s" % (prefix, pull_languages))
                msg = prefix + " translations for resource %s (source: %s)"
                if branch:
                    msg += " for branch {}".format(branch)
                if xliff:
                    msg += " [xliff format]"
                logger.info(msg % (resource, sfile))

            if xliff:
                params.update({'file': 'xliff'})

            for lang in pull_languages:
                local_lang = lang
                if lang in list(lang_map.values()):
                    remote_lang = lang_map.flip[lang]
                else:
                    remote_lang = lang
                if languages and lang not in pull_languages:
                    logger.debug("Skipping language %s" % lang)
                    continue
                if lang != slang:
                    local_file = files.get(lang, None) or files[lang_map[lang]]
                else:
                    local_file = sfile
                logger.debug("Using file %s" % local_file)

                kwargs = {
                    'lang': remote_lang,
                    'stats': stats,
                    'local_file': local_file,
                    'force': force,
                    'mode': mode,
                    'use_git_timestamps': use_git_timestamps
                }

                # xliff files should be always pulled
                if not xliff and not self._should_update_translation(**kwargs):
                    msg = "Skipping '%s' translation (file: %s)."
                    logger.info(
                        msg % (utils.color_text(remote_lang, "RED"),
                               local_file)
                    )
                    continue

                if xliff:
                    local_file += ".xlf"

                if not overwrite:
                    local_file += ".new"

                logger.info(
                    " -> %s: %s" % (utils.color_text(remote_lang, "RED"),
                                    local_file)
                )
                self.do_url_request(
                    url, language=remote_lang, skip_decode=skip_decode,
                    params=params, parallel=parallel, callback=self._save_file,
                    callback_args={"local_file": local_file}
                )

            if new_translations:
                prefix = "Queueing" if parallel else "Pulling"
                msg = prefix + " new translations for resource %s (source: %s)"
                logger.info(msg % (resource, sfile))
                for lang in new_translations:
                    if lang in list(lang_map.keys()):
                        local_lang = lang_map[lang]
                    else:
                        local_lang = lang
                    remote_lang = lang
                    if file_filter:
                        local_file = os.path.relpath(
                            os.path.join(
                                self.root, native_path(
                                    file_filter.replace('<lang>', local_lang)
                                )
                            ), os.curdir
                        )
                    else:
                        trans_dir = os.path.join(self.root, ".tx", resource)
                        if not os.path.exists(trans_dir):
                            os.mkdir(trans_dir)
                        local_file = os.path.relpath(
                            os.path.join(trans_dir,
                                         '%s_translation' % local_lang,
                                         os.curdir))

                    if xliff:
                        local_file += ".xlf"

                    if lang != slang:
                        satisfies_min = self._satisfies_min_translated(
                            stats[remote_lang], mode
                        )
                        if not satisfies_min:
                            msg = "Skipping language %s due to used options."
                            logger.info(msg % lang)
                            continue
                    logger.warning(
                        " -> %s: %s" % (utils.color_text(remote_lang, "RED"),
                                        local_file)
                    )

                    self.do_url_request(
                        url, language=remote_lang, skip_decode=skip_decode,
                        params=params, parallel=parallel,
                        callback=self._save_file,
                        callback_args={"local_file": local_file}
                    )

        # Pull the queued resources
        if parallel:
            logger.info("Pulling resources...")
            perform_parallel_requests()

    def push(self, source=False, translations=False, force=False,
             resources=None, languages=None, skip=False, no_interactive=False,
             xliff=False, branch=None, parallel=False, use_git_timestamps=False):
        """Push all the resources"""
        languages = languages or []
        resources = resources or []

        resource_list = self.get_chosen_resources(resources)
        self.skip = skip
        self.force = force

        params = {}
        if xliff:
            params.update({'file_type': 'xliff'})

        for resource in resource_list:
            push_languages = []
            project_slug, resource_slug = resource.split('.', 1)
            if branch:
                resource_slug = self._slug_with_branch(resource_slug, branch)
            files = self.get_resource_files(resource, xliff=xliff)
            slang = self.get_resource_option(resource, 'source_lang')
            sfile = self.get_source_file(resource)
            lang_map = self.get_resource_lang_mapping(resource)
            host = self.get_resource_host(resource)
            logger.debug("Language mapping is: %s" % lang_map)
            logger.debug("Using host %s" % host)
            self._set_url_info(host=host, project=project_slug,
                               resource=resource_slug)

            message = "Pushing resource {}".format(resource)
            if parallel:
                message = "Resource {} queued for push".format(resource)
            if branch:
                message += " for branch {}".format(branch)
            logger.info(message)

            stats = self._get_stats_for_resource()

            if force and not no_interactive:
                msg = ("Warning: By using --force, the uploaded files will "
                       "overwrite remote translations, even if they are newer "
                       "than your uploaded files.\nAre you sure you want to "
                       "continue?")

                if not utils.confirm(prompt=msg, default=False):
                    logger.info("Aborting...")
                    return

            if source:
                if sfile is None:
                    logger.error("You don't seem to have a proper source file "
                                 "mapping for resource %s. Try without the "
                                 "--source option or set a source file "
                                 "first and then try again." % resource)
                    continue
                # Push source file
                try:
                    push_msg = "Pushing source file (%s)"
                    if parallel:
                        push_msg = "Will queue source file (%s) for push"
                    logger.info(push_msg % sfile)
                    if not self._resource_exists(stats):
                        logger.info("Resource does not exist.  Creating...")
                        fileinfo = "%s;%s" % (resource_slug, slang)
                        filename = self.get_full_path(sfile)
                        self._create_resource(
                            resource, project_slug, fileinfo, filename,
                            branch=branch
                        )
                    self.do_url_request(
                        'push_source', multipart=True, method="PUT",
                        files=[("%s;%s" % (resource_slug, slang),
                                self.get_full_path(sfile)
                                )],
                        params=params,
                        parallel=parallel,
                    )
                except Exception as e:
                    if isinstance(e, SSLError):
                        raise
                    elif not skip:
                        logger.error("Could not upload source file. "
                                     "You can use --skip to ignore this "
                                     "error and continue the execution.")
                        raise
                    else:
                        logger.error(e)
            else:
                try:
                    self.do_url_request('resource_details')
                except Exception as e:
                    if isinstance(e, HttpNotAuthorized):
                        logger.error("Request is not authorized.")
                        continue
                    if isinstance(e, HttpNotFound):
                        msg = "Resource %s doesn't exist on the server."
                        logger.error(msg % resource)
                        continue
                    if isinstance(e, SSLError) or not skip:
                        raise

            if translations:
                # Check if given language codes exist
                if not languages:
                    push_languages = list(files.keys())
                else:
                    push_languages = []
                    for lang in languages:
                        push_languages.append(lang_map[lang]
                                              if lang in list(lang_map.keys())
                                              else lang)

                logger.debug("Languages to push are %s" % push_languages)

                # Push translation files one by one
                for lang in push_languages:
                    local_lang = lang
                    if lang in list(lang_map.values()):
                        remote_lang = lang_map.flip[lang]
                    else:
                        remote_lang = lang

                    local_file = files.get(local_lang)
                    if not local_file:
                        msg = ("Warning: No local translation file found for "
                               "language code '%s'.")
                        logger.error(msg % utils.color_text(lang, "RED"))
                        continue

                    kwargs = {
                        'lang': remote_lang,
                        'stats': stats,
                        'local_file': local_file,
                        'force': force,
                        'use_git_timestamps': use_git_timestamps,
                    }
                    if not self._should_push_translation(**kwargs):
                        msg = "Skipping '%s' translation (file: %s)."
                        logger.info(msg % (utils.color_text(lang, "RED"),
                                    local_file)
                                    )
                        continue

                    msg = "Pushing '%s' translations (file: %s)"
                    if parallel:
                        msg = "Queueing '%s' translations for push (file: %s)"
                    logger.warning(
                        msg % (utils.color_text(remote_lang, "RED"),
                               local_file)
                    )
                    try:
                        self.do_url_request(
                            'push_translation', multipart=True, method='PUT',
                            files=[("%s;%s" % (resource_slug, remote_lang),
                                    self.get_full_path(local_file)
                                    )], language=remote_lang,
                            params=params,
                            parallel=parallel,
                        )
                        logger.debug("Translation %s pushed." % remote_lang)
                    except utils.HttpNotFound:
                        # If the resource exists, we suppose the language code
                        # was invalid
                        if self._resource_exists(stats):
                            error_msg = ("Error: '%s' is not a valid"
                                         " language code. Please refer to"
                                         " https://www.transifex.com/"
                                         "languages/ for a list of supported"
                                         " languages in Transifex." % lang)
                        else:
                            error_msg = ("Resource hasn't been created. "
                                         "Try pushing source file.")
                        logger.error(error_msg)
                    except Exception as e:
                        if isinstance(e, SSLError):
                            raise
                        elif not skip:
                            logger.error("Could not push translations. "
                                         "You can use --skip to ignore this "
                                         "error and continue the execution.")
                            raise
                        else:
                            logger.error(e)

        # Push the queued resources
        if parallel:
            logger.info("Pushing resources...")
            perform_parallel_requests()

    def delete(self, resources=None, languages=None, skip=False, force=False):
        """Delete translations."""
        languages = languages or []
        resources = resources or []

        resource_list = self.get_chosen_resources(resources)
        self.skip = skip
        self.force = force

        if not languages:
            delete_func = self._delete_resource
        else:
            delete_func = self._delete_translations

        for resource in resource_list:
            project_slug, resource_slug = resource.split('.', 1)
            host = self.get_resource_host(resource)
            self._set_url_info(host=host, project=project_slug,
                               resource=resource_slug)
            logger.debug("URL data are: %s" % self.url_info)
            try:
                json, _ = self.do_url_request('project_details', project=self)
            except Exception as e:
                if isinstance(e, HttpNotAuthorized):
                    logger.error("Request is not authorized.")
                    continue
                if isinstance(e, HttpNotFound):
                    msg = "Resource %s doesn't exist on the server."
                    logger.error(msg % resource)
                    continue
                if isinstance(e, SSLError) or not skip:
                    raise

            project_details = utils.parse_json(json)
            stats = self._get_stats_for_resource()
            delete_func(project_details, resource, stats, languages)

    def _delete_resource(self, project_details, resource, stats, *args):
        """Delete a resource from Transifex."""
        project_slug, resource_slug = resource.split('.', 1)
        project_resource_slugs = [
            r['slug'] for r in project_details['resources']
        ]
        logger.info("Deleting resource %s:" % resource)
        if resource_slug not in project_resource_slugs:
            if not self.skip:
                msg = "Skipping: %s : Resource does not exist."
                logger.info(msg % resource)
            return
        if not self.force:
            slang = self.get_resource_option(resource, 'source_lang')
            for language in stats:
                if language == slang:
                    continue
                if int(stats[language]['translated_entities']) > 0:
                    msg = (
                        "Skipping: %s : Unable to delete resource because it "
                        "has a not empty %s translation.\nPlease use -f or "
                        "--force option to delete this resource."
                    )
                    logger.info(msg % (resource, language))
                    return
        try:
            self.do_url_request('delete_resource', method="DELETE")
            self.config.remove_section(resource)
            self.save()
            msg = "Deleted resource %s of project %s."
            logger.info(msg % (resource_slug, project_slug))
        except Exception as e:
            msg = "Unable to delete resource %s of project %s."
            logger.error(msg % (resource_slug, project_slug))
            if isinstance(e, SSLError) or not self.skip:
                raise

    def _delete_translations(self, project_details,
                             resource, stats, languages):
        """Delete the specified translations for the specified resource."""
        logger.info("Deleting translations from resource %s:" % resource)
        for language in languages:
            self._delete_translation(
                project_details, resource, stats, language
            )

    def _delete_translation(self, project_details, resource, stats, language):
        """Delete a specific translation from the specified resource."""
        project_slug, resource_slug = resource.split('.', 1)
        if language not in stats:
            if not self.skip:
                msg = "Skipping %s: Translation does not exist."
                logger.warning(msg % (language))
            return
        if not self.force:
            teams = project_details['teams']
            if language in teams:
                msg = (
                    "Skipping %s: Unable to delete translation because it is "
                    "associated with a team.\nPlease use -f or --force option "
                    "to delete this translation."
                )
                logger.warning(msg % language)
                return
            if int(stats[language]['translated_entities']) > 0:
                msg = (
                    "Skipping %s: Unable to delete translation because it "
                    "is not empty.\nPlease use -f or --force option to delete "
                    "this translation."
                )
                logger.warning(msg % language)
                return
        try:
            self.do_url_request(
                'delete_translation', language=language, method="DELETE"
            )
            msg = "Deleted %s translations of resource %s of project %s."
            logger.info(msg % (language, resource_slug, project_slug))
        except Exception as e:
            msg = "Unable to delete translation %s"
            logger.error(msg % language)
            if isinstance(e, SSLError) or not self.skip:
                raise

    def do_url_request(self, api_call, multipart=False, data=None,
                       files=None, method="GET", skip_decode=False,
                       params=None, parallel=False, no_interactive=False,
                       **kwargs):
        """Issues a url request."""
        files = files or []
        params = params or {}

        # Read the credentials from the config file (.transifexrc)
        host = self.url_info['host']
        username, passwd = self.getset_host_credentials(
            host, no_interactive=no_interactive
        )
        try:
            hostname = self.txrc.get(host, 'hostname')
        except configparser.NoSectionError:
            raise TransifexrcConfigFileError(
                "No entry found for host %s. Edit"
                " ~/.transifexrc and add the appropriate"
                " info in there." % host
            )

        # Create the Url
        kwargs['hostname'] = hostname
        kwargs.update(self.url_info)
        url = API_URLS[api_call] % kwargs

        # in case of GET we need to add xliff option as get parameter
        if params and method == 'GET':
            # update url params
            # in case we need to add extra params on a url, we first get the
            # already existing query, create a dict which will be merged with
            # the extra params and finally put it back in the url
            url_parts = list(urlparse.urlparse(url))
            query = dict(urlparse.parse_qsl(url_parts[4]))
            query.update(params)
            url_parts[4] = urlencode(query)
            url = urlparse.urlunparse(url_parts)

        if multipart:
            for info, filename in files:
                # FIXME: It works because we only pass to files argument
                # only one item
                name = os.path.basename(filename)
                data = {
                    "resource": info.split(';')[0],
                    "language": info.split(';')[1],
                    "uploaded_file": (name, open(filename, 'rb').read())
                }
                # in case of PUT we add xliff option as form data
                if method == 'PUT':
                    data.update(params)

        # Prepare the callback function and arguments
        cb = kwargs.get("callback", None)
        args = kwargs.get("callback_args", {})

        if parallel:
            return utils.queue_request(method, hostname, url, username, passwd,
                                       data, skip_decode=skip_decode,
                                       callback=cb, callback_args=args)

        return utils.make_request(
            method, hostname, url, username, passwd, data,
            skip_decode=skip_decode, callback=cb, callback_args=args
        )

    def _should_update_translation(self, lang, stats, local_file, force=False,
                                   mode=None, use_git_timestamps=False):
        """Whether a translation should be udpated from Transifex.

        We use the following criteria for that:
        - If user requested to force the download.
        - If language exists in Transifex.
        - If the local file is older than the Transifex's file.
        - If the user requested a x% completion.

        Args:
            lang: The language code to check.
            stats: The (global) statistics object.
            local_file: The local translation file.
            force: A boolean flag.
            mode: The mode for the translation.
            use_git_timestamps: Boolean flag -  use latest commit timestamp
                instead of system timestamps for checking if local file is
                older than Transifex's.
        Returns:
            True or False.
        """
        return self._should_download(lang, stats, local_file, force, use_git_timestamps=use_git_timestamps)

    def _should_add_translation(self, lang, stats, force=False, mode=None, use_git_timestamps=False):
        """Whether a translation should be added from Transifex.

        We use the following criteria for that:
        - If user requested to force the download.
        - If language exists in Transifex.
        - If the user requested a x% completion.

        Args:
            lang: The language code to check.
            stats: The (global) statistics object.
            force: A boolean flag.
            mode: The mode for the translation.
            use_git_timestamps: Boolean flag -  use latest commit timestamp
                instead of system timestamps for checking if local file is
                older than Transifex's.
        Returns:
            True or False.
        """
        return self._should_download(lang, stats, None, force, use_git_timestamps=use_git_timestamps)

    def _should_download(self, lang, stats, local_file=None, force=False,
                         mode=None, use_git_timestamps=False):
        """Return whether a translation should be downloaded.

        If local_file is None, skip the timestamps check (the file does
        not exist locally).
        """
        try:
            lang_stats = stats[lang]
        except KeyError:
            logger.debug("No lang %s in statistics" % lang)
            return False

        satisfies_min = self._satisfies_min_translated(lang_stats, mode)
        if not satisfies_min:
            return False

        if force:
            logger.debug("Downloading translation due to -f")
            return True

        if local_file is not None:
            remote_update = self._extract_updated(lang_stats)
            if not self._remote_is_newer(remote_update, local_file, use_git_timestamps):
                logger.debug("Local is newer than remote for lang %s" % lang)
                return False
        return True

    def _should_push_translation(self, lang, stats, local_file, force=False, use_git_timestamps=False):
        """Return whether a local translation file should be
        pushed to Trasnifex.

        We use the following criteria for that:
        - If user requested to force the upload.
        - If language exists in Transifex.
        - If local file is younger than the remote file.

        Args:
            lang: The language code to check.
            stats: The (global) statistics object.
            local_file: The local translation file.
            force: A boolean flag.
            use_git_timestamps: Boolean flag -  use latest commit timestamp
                instead of system timestamps for checking if local file is
                older than Transifex's.
        Returns:
            True or False.
        """
        if force:
            logger.debug("Push translation due to -f.")
            return True
        try:
            lang_stats = stats[lang]
        except KeyError:
            logger.debug("Language %s does not exist in Transifex." % lang)
            return True
        if local_file is not None:
            remote_update = self._extract_updated(lang_stats)
            if self._remote_is_newer(remote_update, local_file, use_git_timestamps):
                msg = "Remote translation is newer than local file for lang %s"
                logger.debug(msg % lang)
                return False
        return True

    def _generate_timestamp(self, update_datetime):
        """Generate a UNIX timestamp from the argument.

        Args:
            update_datetime: The datetime in the format used by Transifex.
        Returns:
            A float, representing the timestamp that corresponds to the
            argument.
        """
        time_format = "%Y-%m-%d %H:%M:%S"
        return time.mktime(
            datetime.datetime(
                *time.strptime(update_datetime, time_format)[0:5]
            ).utctimetuple()
        )

    def _get_time_of_local_file(self, path, use_git_timestamps=False):
        """Get the modified time of the path_.

        Args:
            path: The path we want the mtime for.
            use_git_timestamps: Boolean flag -  use latest commit timestamp
                instead of system timestamps for checking if local file is
                older than Transifex's.
        Returns:
            The time as a timestamp or None, if the file does not exist
        """
        if not os.path.exists(path):
            return None

        if use_git_timestamps:
            epoch_timestamp = utils.get_git_file_timestamp(path)
            if epoch_timestamp:
                return time.mktime(time.gmtime(epoch_timestamp))
            else:
                logger.warning(
                    "Failed to find git repository. Fallback to OS timstamp"
                )

        return time.mktime(time.gmtime(os.path.getmtime(path)))

    def _satisfies_min_translated(self, stats, mode=None):
        """Check whether a translation fulfills the filter used for
        minimum translated percentage.

        Args:
            perc: The current translation percentage.
        Returns:
            True or False
        """
        cur = self._extract_completed(stats, mode)
        option_name = 'minimum_perc'
        if self.minimum_perc is not None:
            minimum_percent = self.minimum_perc
        else:
            global_minimum = int(
                self.get_resource_option('main', option_name) or 0
            )
            resource_minimum = int(
                self.get_resource_option(
                    self.resource, option_name
                ) or global_minimum
            )
            minimum_percent = resource_minimum
        return cur >= minimum_percent

    def _remote_is_newer(self, remote_updated, local_file, use_git_timestamps=False):
        """Check whether the remote translation is newer that the local file.

        Args:
            remote_updated: The date and time the translation was last
                updated remotely.
            local_file: The local file.
            use_git_timestamps: Boolean flag -  use latest commit timestamp
                instead of system timestamps for checking if local file is
                older than Transifex's.
        Returns:
            True or False.
        """
        if remote_updated is None:
            logger.debug("No remote time")
            return False
        remote_time = self._generate_timestamp(remote_updated)
        local_time = self._get_time_of_local_file(
            self.get_full_path(local_file),
            use_git_timestamps
        )
        logger.debug(
            "Remote time is %s and local %s" % (remote_time, local_time)
        )
        if local_time is not None and remote_time < local_time:
            return False
        return True

    @classmethod
    def _extract_completed(cls, stats, mode=None):
        """Extract the information for the translated percentage from the stats.

        Args:
            stats: The stats object for a language as returned by Transifex.
            mode: The mode of translations requested.
        Returns:
            The percentage of translation as integer.
        """
        if mode == 'reviewed':
            key = 'reviewed_percentage'
        else:
            key = 'completed'
        try:
            return int(stats[key][:-1])
        except KeyError:
            return 0

    @classmethod
    def _extract_updated(cls, stats):
        """Extract the  information for the last update of a translation.

        Args:
            stats: The stats object for a language as returned by Transifex.
        Returns:
            The last update field.
        """
        try:
            return stats['last_update']
        except KeyError:
            return None

    def _download_pseudo(self, project_slug, resource_slug, pseudo_file):
        response, charset = self.do_url_request(
            'pull_pseudo_file',
            resource_slug=resource_slug,
            project_slug=project_slug
        )
        response = utils.parse_json(response)

        base_dir = os.path.split(pseudo_file)[0]
        utils.mkdir_p(base_dir)

        with open(pseudo_file, "wb") as fd:
            fd.write(response['content'].encode("utf-8"))

    def _new_translations_to_add(self, files, slang, lang_map,
                                 stats, force=False, use_git_timestamps=False):
        """Return a list of translations which are
        new to the local installation.
        """
        new_translations = []
        langs = list(stats.keys())
        logger.debug("Available languages are: %s" % langs)

        for lang in langs:
            lang_exists = lang in list(files.keys())
            lang_is_source = lang == slang
            mapped_lang_exists = (
                lang in lang_map and lang_map[lang] in list(files.keys())
            )
            if lang_exists or lang_is_source or mapped_lang_exists:
                continue
            if self._should_add_translation(lang, stats, force, use_git_timestamps):
                new_translations.append(lang)
        return set(new_translations)

    def _get_stats_for_resource(self, no_interactive=False):
        """Get the statistics information for a resource."""
        try:
            r, charset = self.do_url_request(
                'resource_stats', no_interactive=no_interactive
            )
            logger.debug("Statistics response is %s" % r)
            stats = utils.parse_json(r)
        except utils.HttpNotFound:
            logger.debug("Resource not found, creating...")
            stats = {}
        except Exception as e:
            logger.debug(six.u(str(e)))
            raise
        return stats

    def get_chosen_resources(self, resources):
        """Get the resources the user selected.

        Support wildcards in the resources specified by the user.

        Args:
            resources: A list of resources as specified in command-line or
                an empty list.
        Returns:
            A list of resources.
        """
        configured_resources = self.get_resource_list()
        if not resources:
            return configured_resources

        selected_resources = []
        for resource in resources:
            found = False
            for full_name in configured_resources:
                if fnmatch.fnmatch(full_name, resource):
                    selected_resources.append(full_name)
                    found = True
            if not found:
                msg = "Specified resource '%s' does not exist."
                raise Exception(msg % resource)
        logger.debug("Operating on resources: %s" % selected_resources)
        return selected_resources

    def _languages_to_pull(self, languages, files, lang_map, stats, force):
        """Get a set of languages to pull.

        Args:
            languages: A list of languages the user selected in command.
            files: A dictionary of current local translation files.
        Returns:
            A tuple of a set of existing languages and new translations.
        """
        if not languages:
            pull_languages = set([])
            pull_languages |= set(files.keys())
            mapped_files = []
            for lang in pull_languages:
                if lang in lang_map.flip:
                    mapped_files.append(lang_map.flip[lang])
            pull_languages -= set(lang_map.flip.keys())
            pull_languages |= set(mapped_files)
            return (pull_languages, set([]))
        else:
            pull_languages = []
            new_translations = []
            f_langs = list(files.keys())
            for l in languages:
                if l not in f_langs and not (
                        l in lang_map and lang_map[l] in f_langs):
                    if self._should_add_translation(l, stats, force):
                        new_translations.append(l)
                else:
                    pull_languages.append(lang_map[l]
                                          if l in list(lang_map.keys())
                                          else l)
            return (set(pull_languages), set(new_translations))

    def _extension_for(self, i18n_type):
        """Return the extension used for the specified type."""
        try:
            json, charset = self.do_url_request('formats')
            res = utils.parse_json(json)
            return res[i18n_type]['file-extensions'].split(',')[0]
        except Exception as e:
            logger.warning(
                "The file extension for i18n_type %s is not found." % str(e)
            )
            return ''

    def _resource_exists(self, stats):
        """Check if resource exists.

        Args:
            stats: The statistics dict as returned by Tx.
        Returns:
            True, if the resource exists in the server.
        """
        return bool(stats)

    def _create_resource(self, resource, pslug, fileinfo, filename, **kwargs):
        """Create a resource.

        Args:
            resource: The full resource name.
            pslug: The slug of the project.
            fileinfo: The information of the resource.
            filename: The name of the file.
        Raises:
            URLError, in case of a problem.
        """
        method = "POST"
        api_call = 'create_resource'

        host = self.url_info['host']
        try:
            hostname = self.txrc.get(host, 'hostname')
            username, passwd = self.getset_host_credentials(host)
        except configparser.NoSectionError:
            raise TransifexrcConfigFileError(
                "No user credentials found for host %s. Edit "
                "~/.transifexrc and add the appropriate "
                "info in there." % host)

        # Create the Url
        kwargs['hostname'] = hostname
        kwargs.update(self.url_info)
        kwargs['project'] = pslug
        url = (API_URLS[api_call] % kwargs)

        i18n_type = self._get_option(resource, 'type')
        if i18n_type is None:
            raise ConfigFileError(
                "Please define the resource type in "
                ".tx/config (eg. type = PO). "
                "More info: http://bit.ly/txcconfig"
            )

        name = os.path.basename(filename)
        data = {
            "slug": fileinfo.split(';')[0],
            "name": fileinfo.split(';')[0],
            "uploaded_file": (name, open(filename, 'rb').read()),
            "i18n_type": i18n_type
        }
        if kwargs.get('branch'):
            data.update({'category': kwargs['branch']})

        r, charset = utils.make_request(
            method, hostname, url, username, passwd, data
        )
        return r

    def _get_url_by_pull_mode(self, mode):
        """Get the URL by the pull mode.

        If the pull mode is not valid, the default pull mode will be used.
        """
        if mode is not None:
            try:
                # check if specific url name exists in the url list
                API_URLS[PULL_MODE_URL_NAME.format(mode=mode)]
                return PULL_MODE_URL_NAME.format(mode=mode)
            except KeyError:
                logger.warning('Invalid mode provided. ' +
                               'Default pull mode will be used')

        return DEFAULT_PULL_URL

    def _get_option(self, resource, option):
        """Get the value for the option in the configuration file.

        If the option is not in the resource section, look for it in
        the project.

        Args:
            resource: The resource name.
            option: The option the value of which we are interested in.
        Returns:
            The option value or None, if it does not exist.
        """
        value = self.get_resource_option(resource, option)
        if value is None:
            if self.config.has_option('main', option):
                return self.config.get('main', option)
        return value

    def set_i18n_type(self, resources, i18n_type):
        """Set the type for the specified resources."""
        self._set_resource_option(resources, key='type', value=i18n_type)

    def set_min_perc(self, resources, perc):
        """Set the minimum percentage for the resources."""
        self._set_resource_option(resources, key='minimum_perc', value=perc)

    def set_default_mode(self, resources, mode):
        """Set the default mode for the specified resources."""
        self._set_resource_option(resources, key='mode', value=mode)

    def _set_resource_option(self, resources, key, value):
        """Set options in the configuration file.

        If resources is empty. set the option globally.
        """
        if not resources:
            self.config.set('main', key, value)
            return
        for r in resources:
            self.config.set(r, key, value)

    @staticmethod
    def _save_file(local_file="", charset=None, data=None):
        base_dir = os.path.split(local_file)[0]
        utils.mkdir_p(base_dir)
        fd = open(local_file, 'wb')
        if charset is not None:
            data = data.encode(charset)
        fd.write(data)
        fd.close()

    def _set_url_info(self, host, project, resource):
        self.url_info = {
            'host': host,
            'project': project,
            'resource': resource
        }



try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import six

from txclib.exceptions import MalformedConfigFile


class OrderedRawConfigParser(configparser.RawConfigParser):
    """
    Overload standard Class ConfigParser.RawConfigParser
    """
    def write(self, fp):
        """Write an .ini-format representation of the configuration state."""
        if self._defaults:
            fp.write("[%s]\n" % DEFAULTSECT)
            for key in sorted(self._defaults):
                fp.write("%s = %s\n" % (key, str(self._defaults[key]).
                         replace('\n', '\n\t')))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for key in sorted(self._sections[section]):
                if key != "__name__":
                    fp.write("%s = %s\n" %
                             (key, str(self._sections[section][key]).
                              replace('\n', '\n\t')))
            fp.write("\n")

    optionxform = str


_NOTFOUND = object()


class Flipdict(dict):
    """An injective (one-to-one) python dict.  Ensures that each key maps
    to a unique value, and each value maps back to that same key.

    Code mostly taken from here:
    http://code.activestate.com/recipes/576968-flipdict-python-dict-that-also-maintains-a-one-to-/
    """

    def __init__(self, *args, **kw):
        self._flip = dict.__new__(self.__class__)
        setattr(self._flip, "_flip", self)
        for key, val in six.iteritems(dict(*args, **kw)):
            self[key] = val

    @property
    def flip(self):
        """The inverse mapping."""
        return self._flip

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, dict(self))

    __str__ = __repr__

    def copy(self):
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, keys, value=None):
        return cls(dict.fromkeys(keys, value))

    def __setitem__(self, key, val):
        k = self._flip.get(val, _NOTFOUND)
        if not (k is _NOTFOUND or k == key):
            raise MalformedConfigFile("Your lang map configuration is not correct. "
                                      "Duplicate entry detected with value '%s'. "
                                      "Keys and values should be unique." % val)

        v = self.get(key, _NOTFOUND)
        if v is not _NOTFOUND:
            dict.__delitem__(self._flip, v)

        dict.__setitem__(self,       key, val)
        dict.__setitem__(self._flip, val, key)

    def setdefault(self, key, default=None):
        # Copied from python's UserDict.DictMixin code.
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def update(self, other=None, **kwargs):
        # Copied from python's UserDict.DictMixin code.
        # Make progressively weaker assumptions about "other"
        if other is None:
            pass
        elif hasattr(other, 'items'):
            for k, v in six.iteritems(other):
                self[k] = v
        elif hasattr(other, 'keys'):
            for k in list(other.keys()):
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)

    def __delitem__(self, key):
        val = dict.pop(self, key)
        dict.__delitem__(self._flip, val)

    def pop(self, key, *args):
        val = dict.pop(self, key, *args)
        dict.__delitem__(self._flip, val)
        return val

    def popitem(self):
        key, val = dict.popitem(self)
        dict.__delitem__(self._flip, val)
        return key, val

    def clear(self):
        dict.clear(self)
        dict.clear(self._flip)


import os
import ssl

CERT_REQUIRED = getattr(ssl, os.environ.get('TX_CERT_MODE', 'CERT_REQUIRED'))
